/*
 *  ofxSndObj.cpp
 *  ofxSndObj
 *
 *  Created by joshua noble on 10/5/08.
 *  thefactoryfactory
 *
 */

#include "ofxSndObj.h"


ofxSndObj::ofxSndObj(){
	//srand( (unsigned) time( (time_t *) 0 ));
}


ofxSndObj::~ofxSndObj() 
{
	createdOscilis.clear();
	createdBuzzes.clear();
	createdLoops.clear();
	createdNoises.clear();
	createdTables.clear();
	createdMixers.clear();
	createdSndIns.clear();
	createdFilters.clear();
	createdSpectrums.clear();
}

void ofxSndObj::cleanup()
{
	createdOscilis.clear();
	createdBuzzes.clear();
	createdLoops.clear();
	createdNoises.clear();
	createdTables.clear();
	createdMixers.clear();
	createdSndIns.clear();
	createdFilters.clear();
	createdSpectrums.clear();
}

int ofxSndObj::startSystemInOut(bool useSystemIn, bool useSystemOut)
{
	
#ifdef MACOSX
	
	core = new SndRTIO(2, SND_IO);
	thread.AddObj(core, SNDIO_IN);
	
#else
	
	rtioIn = new SndRTIO(1, SND_INPUT);
	rtioOut = new SndRTIO(2, SND_OUTPUT);
	thread.AddObj(rtioIn, SNDIO_IN);
	
#endif
	
	hasInitedOut = true;
	hasInitedIn = true;
	
	return 1;
}


int ofxSndObj::addObject(SndObj* obj)
{
	
	addedObjs.push_back((SndObj*)obj);
	thread.AddObj((SndObj*)obj);
	
	return 1;
}

int ofxSndObj::stopProcessing()
{	
	thread.ProcOff();
}

int ofxSndObj::startProcessing()
{	
	thread.ProcOn();
}


int ofxSndObj::setOutput(ofxBaseSndObj outputObj, int intype, int channel)
{
	SndObj* so = findSndObj(outputObj, intype);
	
#ifdef MACOSX
	core->SetOutput(channel, so);
#else
	rtioIn->SetOutput(channel, so);
#endif
}


int ofxSndObj::setOutput(SndObj* obj, int channel)
{
#ifdef MACOSX
	core->SetOutput(channel, obj);
#else
	rtioIn->SetOutput(channel, obj);
#endif
}

int ofxSndObj::startOut(bool useSystemOut, int device = 1)
{
#ifdef MACOSX
	core = new SndRTIO(2, SND_OUTPUT, 512);
	thread.AddObj(core, SNDIO_OUT);
#else
	rtioOut = new SndRTIO(2, SND_OUTPUT);
	thread.AddObj(rtioOut, SNDIO_OUT);
#endif	
	hasInitedOut = true;
	return 1;
}

string ofxSndObj::createTable(long length, int type, float phase)
{
	HarmTable* tbl = new HarmTable(length, SINE, phase);
	string key = generateRandomStringKey();
	createdTables[key] = (Table*)tbl;
	return key;
}
void ofxSndObj::removeTable(string key)
{
	HarmTable* tbl = (HarmTable*) createdTables[key];
	thread.DeleteObj((SndObj*)tbl);
	delete createdOscilis[key];
}
void ofxSndObj::tuneTable(string key, float frequency)
{
	HarmTable* tbl = (HarmTable*) createdTables[key];
	tbl->SetPhase(frequency);
}

string ofxSndObj::createOscilator(ofxSOTable t, float frq, float amp)
{
	Table* tbl = createdTables[t.key];
	Oscili* o = new Oscili(tbl, frq, amp);
	string key = generateRandomStringKey();
	createdOscilis[key] = o;
	thread.AddObj(o);
	return key;
	
}

string ofxSndObj::createOscilator(ofxSOTable t, float frq, float amp, ofxBaseSndObj b, int type){

	Table* tbl = createdTables[t.key];
	
	SndObj* so;
	
	if(type == OFXSNDOBJOSCILATOR) {
		so = createdOscilis[b.key];
	} else if(type == OFXSNDOBJBUZZ) {
		so = createdBuzzes[b.key];
	}
	Oscili* o = new Oscili(tbl, frq, amp, so);
	string key = generateRandomStringKey();
	createdOscilis[key] = o;
	thread.AddObj(o);
	return key;
	
}

string ofxSndObj::createBuzz(float fr, float amp, int harm){

	Buzz* b = new Buzz(fr, amp, harm);
	string key = generateRandomStringKey();
	createdBuzzes[key] = b;
	thread.AddObj(b);
	return key;
}


void ofxSndObj::removeOscilator(string key)
{
	Oscili* o = createdOscilis[key];
	thread.DeleteObj(o);
	delete createdOscilis[key];
}


void ofxSndObj::removeBuzz(string key)
{
	Buzz* b = createdBuzzes[key];
	thread.DeleteObj(b);
	delete createdBuzzes[key];
}


void ofxSndObj::tuneOscilator(string key, float frq = 0, float amp = 0)
{
	Oscili* o = createdOscilis[key];
	if(frq != 0)
	{
		o->SetFreq(frq, NULL);
	}
	if(amp != 0)
	{
		o->SetAmp(0, NULL);
	}

}

void ofxSndObj::tuneBuzz(string key, float frequency = 0, float amp = 0){
	Buzz* b = createdBuzzes[key];
	if(frequency != 0)
	{
		b->SetFreq(frequency, 0);
	}
	if(amp != 0)
	{
		b->SetAmp(amp, NULL);
	}
}

string ofxSndObj::createNoise(float amp){

	Rand* r = new Rand(amp);
	string key = generateRandomStringKey();
	createdNoises[key] = r;
	thread.AddObj(r);
	return key;
	

}

void ofxSndObj::removeNoise(string key)
{
	Rand* r = createdNoises[key];
	thread.DeleteObj(r);
	delete createdNoises[key];

}

void ofxSndObj::tuneNoise(string key, float amp)
{
	Rand* r = createdNoises[key];
	r->SetAmp(amp);
	delete createdNoises[key];

}

string ofxSndObj::createMixer()
{	
	Mixer* r = new Mixer();
	string key = generateRandomStringKey();
	createdMixers[key] = r;
	thread.AddObj(r);
	
	setOutput(r, 1);
	setOutput(r, 2);
	
	
	return key;
	
}

void ofxSndObj::addToMixer(string key, ofxBaseSndObj added, int type)
{
	Mixer* m = createdMixers[key];
	
	SndObj* so = findSndObj(added, type);

	m->AddObj(so);
}

void ofxSndObj::addToMixer(ofxSOMixer mix, ofxBaseSndObj added, int type )
{
	Mixer* m = createdMixers[mix.key];
	m->AddObj(findSndObj(added, type));
	
}

void ofxSndObj::removeFromMixer(string key, ofxBaseSndObj base, int type )
{
	Mixer* m = createdMixers[key];
	m->DeleteObj(findSndObj(base, type));

	// more as more types
}


string ofxSndObj::createSpecMult(string type, ofxBaseSndObj inObj1,  int obj1Type, ofxBaseSndObj inObj2, int obj2Type)
{
	
	SndObj* so1 = findSndObj(inObj1, obj1Type);
	SndObj* so2 = findSndObj(inObj2, obj2Type);
	
	SpecMult* sm = new SpecMult(so1, so2);
	string key = generateRandomStringKey();
	createdSpectrums[key] = sm;
	thread.AddObj(sm);
	return key;
	
}

void ofxSndObj::removeSpecMult(string key)
{
	SpecMult* sm = createdSpectrums[key];
	thread.DeleteObj(sm);
	delete createdMixers[key];
}

string ofxSndObj::createFilter(string filterType, float frequency, float bandwidth, ofxBaseSndObj inObject, int type)
{
	SndObj* so = findSndObj(inObject, type);
	genFilter(filterType, frequency, bandwidth, so);
}

void ofxSndObj::genFilter(string filterType, float frequency, float bandwidth, SndObj* inObject)
{
	Filter* f;
	const char* fil = filterType.c_str();
	if(strcmp(fil, "Resonator") == 0) {
		f = new Reson(frequency, bandwidth, inObject);
	} else if(strcmp(fil, "HiPass") == 0) {
		f = new HiPass(frequency, inObject);
	}else if(strcmp(fil, "LoPass") == 0) {
		f = new LoPass(frequency, inObject);
	} else if(strcmp(fil, "AllPass") == 0) {
		f = new Ap(frequency, bandwidth, inObject);
	} else if(strcmp(fil, "ButtHP") == 0) {
		f = new ButtHP(frequency, inObject);
	} else if(strcmp(fil, "ButtLP") == 0) {
		f = new ButtLP(frequency, inObject);
	}
	
	thread.AddObj(f);
}

void ofxSndObj::removeFilter(string key)
{

}
void ofxSndObj::tuneFilter(string key, float frequency, float bandwidth)
{}


string ofxSndObj::createPluck(float fr, float amp, float fdbgain)
{
	Pluck* p = new Pluck(fr, amp, fdbgain);
	thread.AddObj(p);
	string s = generateRandomStringKey();
	createdPlucks[s] = p;
	return s;

}

string ofxSndObj::createPluck(float fr, float amp, float fdbgain, ofxBaseSndObj b, int type)
{
	
	Pluck* p;
	SndObj* so = findSndObj(b, type);

	p = new Pluck(fr, amp, fdbgain, so);
	thread.AddObj(p);
	string s = generateRandomStringKey();
	createdPlucks[s] = p;
	return s;
}

void ofxSndObj::removePluck(string key)
{
	thread.DeleteObj(createdPlucks[key]);
	delete createdPlucks[key];
}
void ofxSndObj::tunePluck(string key, float amp)
{
	createdPlucks[key]->SetAmp(amp);
}

void ofxSndObj::pluck(string key)
{
	createdPlucks[key]->RePluck();
}

//float att, float maxamp, float dec, float sus, float rel, float dur, SndObj* InObj = 0,
string ofxSndObj::createADSR(float attack, float maxamp, float decay, float sustain, float release, float duration)
{
	ADSR* adsr = new ADSR(attack, maxamp, decay, sustain, release, duration);
	thread.AddObj(adsr);
	string s = generateRandomStringKey();
	createdADSRs[s] = adsr;
	return s;
}

string ofxSndObj::createADSR(float attack, float maxamp, float decay, float sustain, float release, float duration, ofxBaseSndObj inObject, int type)
{
	ADSR* adsr;
	SndObj* so = findSndObj(inObject, type);

	adsr = new ADSR(attack, maxamp, decay, sustain, release, duration, so);
	thread.AddObj(adsr);
	string s = generateRandomStringKey();
	createdADSRs[s] = adsr;
	return s;
}

void ofxSndObj::adsrSustain(string key)
{
	createdADSRs[key]->Sustain();
}

void ofxSndObj::tuneADSRMaxAmp(string key, float maxamp)
{
	createdADSRs[key]->SetMaxAmp(maxamp);
}
void ofxSndObj::tuneADSR(string key, float att, float dec, float sus, float rel)
{
	createdADSRs[key]->SetADSR(att, dec, sus, rel);
}
void ofxSndObj::tuneADSRDuration(string key, float dur)
{
	createdADSRs[key]->SetDur(dur);
}

//(float xfadetime, float looptime, SndObj* InObj, float pitch = 1.f,  int vecsize=DEF_VECSIZE, float sr=DEF_SR
string ofxSndObj::createLoop(float xfadetime, float looptime, ofxBaseSndObj inObject, int intype, float pitch = 1.f)
{
	SndLoop* sl;
	SndObj* so = findSndObj(inObject, intype);
	
	sl = new SndLoop(xfadetime, looptime, so, pitch);
	thread.AddObj(sl);
	
	string key = generateRandomStringKey();
	createdLoops[key] = sl;
	
	return key;
}

void ofxSndObj::deleteLoop(string key)
{
	SndLoop* sl = createdLoops[key];
	thread.DeleteObj(sl);
	delete createdLoops[key];
}

void ofxSndObj::resampleLoop(string key) {
	SndLoop* sl = createdLoops[key];
	sl->ReSample();
}

string ofxSndObj::loadSound(string filename, bool stereo, int channel)
{
	SndWave* sw = new SndWave((char*) filename.c_str(), READ);//, (stereo) ? 2 : 1, channel);
	
	SndIn* si = new SndIn(sw, 1);
	string key = generateRandomStringKey();
	createdSndWavs[key] = sw;
	createdSndIns[key] = si;
	
	thread.AddObj(sw, SNDIO_IN);
	
	thread.AddObj(si);

	return key;
}

void ofxSndObj::deleteSound(string key)
{
	SndWave* sw = createdSndWavs[key];
	SndIn* si = createdSndIns[key];
	thread.DeleteObj(si);

	delete sw;
	delete si;
}


string ofxSndObj::generateRandomStringKey( )
{
	int len = 20;
	char str[21];
	char *chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
	unsigned int max = strlen( chars );
	unsigned long i = 0L;
	
	for ( ; i < len - 2L; ++i ) {
		str[ i ] = chars[ rand() % max ];
	}
	str[ i++ ] = '\n';
	str[ i ] = '\0';
	
	return str;
}


SndObj* ofxSndObj::findSndObj(ofxBaseSndObj added, int type)
{
	if(type == OFXSNDOBJOSCILATOR)
	{
		return (SndObj*) createdOscilis[added.key];
	}
	if(type == OFXSNDOBJBUZZ)
	{
		return (SndObj*) createdBuzzes[added.key];
	}
	if( type == OFXSNDOBJPLUCK)
	{
		return (SndObj*) createdPlucks[added.key];
	}
	if( type == OFXSNDOBJADSR)
	{
		return (SndObj*)  createdADSRs[added.key];
	}
	if( type == OFXSNDOBJWAV )
	{
		return (SndObj*) createdSndIns[added.key];
	}
	if( type == OFXSNDOBJLOOP )
	{
		return (SndObj*) createdLoops[added.key];
	}
	if( type == OFXSNDOBJMIXER )
	{
		return (SndObj*) createdMixers[added.key];
	}
}
